#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

__all__ = ['Storage', 'Preset', 'preset', 'Splitter', 'Scaler', 'BoxCox', 'MDL', 'Dataset']

from .dataset import Dataset
from .mdl import MDL
from .preset import Preset, preset
from .splitter import Splitter
from .storage import Storage
from .transform import Scaler, BoxCox



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from collections import namedtuple

from torch import nn
from torch.nn import Sequential

from .layer import Layer1d
from .wrap import L1
from ...math import Product


class Generator1d(object):
    """
    Generate random model from the supplied parameters.
    """

    def __init__(self, n_features: int, n_predict: int, *,
                 output_layer=L1.linear(),
                 n_neuron,
                 drop_out=(0.0,),
                 layer_func=(L1.linear(),),
                 act_func=(nn.ReLU(),),
                 batch_normalize=(L1.batch_norm(),)
                 ):
        """
        Parameters
        ----------
        n_features: int
            Input dimension.
        n_predict: int
            Output dimension.
        output_layer: func
            Output layer.
        n_neuron: [int]
            Number of neuron.
        drop_out: [float]
            Dropout rate.
        layer_func: [func]
            Layer functions. such like: :meth:`~.L1.linear`.
        act_func: [func]
            Activation functions. such like: :class:`torch.nn.ReLU`.
        batch_normalize: [bool]
            Batch Normalization. such like: :meth:`~.L1.batch_norm`.
        """
        self.output_layer = output_layer
        self.n_in, self.n_out = n_features, n_predict

        # save parameters
        self.drop_out = drop_out
        self.layer_func = layer_func
        self.act_func = act_func
        self.batch_normalize = batch_normalize

        # calculate layer's variety
        self.layer_var = Product(n_neuron, drop_out, layer_func, act_func, batch_normalize)

    def __call__(self, hidden: int, *, n_models: int = 0, scheduler=None, replace=False):
        """
        Generate sample model.

        Parameters
        ----------
        hidden: int
            Number of hidden layers.
        n_models: int
            Number of model sample
        scheduler: func
            A function be used to determining the layer parameters from previous layer.

            .. py:function:: scheduler(index, paras) -> dict

                index: int
                    Index of  current layer.
                paras: dict
                    Layer parameters Include:
                    ``n_in``, ``n_out``, ``p_drop``, ``layer_func``, ``act_func``, ``batch_nor``.
                return: dict
                    Layer parameters as dict.

        Returns
        -------
        iterable
            Random models as generator.
            Can be access with :func:`next()` or ``for ... in models`` statement.

        Examples
        --------
        >>> from  math import ceil
        >>> from random import uniform
        >>> scheduler = lambda index, pars: dict(paras, n_out=ceil(paras['n_out'] * uniform(0.5, 0.8)))
        """
        from numpy.random import choice
        named_paras = ['n_in', 'n_out', 'drop_out', 'layer_func', 'act_func', 'batch_nor']
        layer = namedtuple('LayerParas', named_paras)

        if scheduler is None:
            all_ = Product(self.layer_var, repeat=hidden)
            all_size = self.layer_var.size ** hidden

            if n_models == 0:
                n_models = all_size
            if n_models > all_size and not replace:
                raise ValueError("larger sample than population({}) when 'replace=False'".format(all_size))

            # sampling all_
            samples = choice(all_size, n_models, replace).tolist()
            for i in samples:
                layer_paras = all_[i]

                # set layers
                layers = list()
                n_in = self.n_in
                for para in layer_paras:
                    layer_ = layer(*((n_in,) + para))._asdict()
                    layers.append(Layer1d(**layer_))
                    n_in = para[0]
                out_layer = self.output_layer(n_in, self.n_out)
                layers.append(out_layer)

                yield Sequential(*layers)

        else:
            if n_models == 0:
                n_models = self.layer_var.size
            if n_models > self.layer_var.size and not replace:
                raise ValueError("larger sample than population({}) when 'replace=False'".format(self.layer_var.size))

            samples = choice(self.layer_var.size, n_models, replace).tolist()

            for i in samples:

                layers = list()
                n_in = self.n_in
                paras = self.layer_var[i]
                layer_ = layer(*((n_in,) + paras))._asdict()

                layers.append(Layer1d(**layer_))
                n_in = layer_['n_out']
                for n in range(hidden - 1):
                    layer_ = scheduler(n + 1, layer_)
                    layer_['n_in'] = n_in
                    layers.append(Layer1d(**layer_))
                    n_in = layer_['n_out']
                out_layer = self.output_layer(n_in, self.n_out)
                layers.append(out_layer)

                yield Sequential(*layers)

        return



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import argparse
from os import remove, rename
from pathlib import Path

from tqdm import tqdm

from .utils import config


def migrate(args_):
    data_path = Path('~/.xenonpy/dataset').expanduser().absolute()
    user_path = Path(config('userdata')).expanduser().absolute()

    def migrate_(f):
        path = data_path / f
        if not path.exists():
            return
        if args_.keep:
            path_ = user_path / f
            rename(str(path), str(path_))
        else:
            remove(str(path))

    for file in tqdm(['mp_inorganic.pkl.pd_', 'mp_structure.pkl.pd_', 'oqmd_inorganic.pkl.pd_',
                      'oqmd_structure.pkl.pd_'], desc='migrating'):
        migrate_(file)


parser = argparse.ArgumentParser(
    prog='XenonPy',
    description='''
    XenonPy is a Python library that implements a comprehensive set of
    machine learning tools for materials informatics.
    ''')
subparsers = parser.add_subparsers()

parser_migrate = subparsers.add_parser('migrate', help='see `migrate -h`')
parser_migrate.add_argument(
    '-k',
    '--keep',
    action='store_true',
    help=
    'Keep files fetched from `yoshida-lab/dataset`. These files will be moved to `userdata` dir.'
)
parser_migrate.set_defaults(handler=migrate)

args = parser.parse_args()
if hasattr(args, 'handler'):
    args.handler(args)
else:
    parser.print_help()



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

__all__ = ['wrap', 'Checker', 'Layer1d', 'RegressionRunner', 'persist', 'BaseRunner']

from .checker import Checker
from .layer import Layer1d
from .model_maker import Generator1d
from .runner import BaseRunner, RegressionRunner, persist



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from pathlib import Path

from ruamel.yaml import YAML

__all__ = [
    '__pkg_name__',
    '__version__',
    '__db_version__',
    '__release__',
    '__short_description__',
    '__license__',
    '__author__',
    '__author_email__',
    '__maintainer__',
    '__maintainer_email__',
    '__github_username__',
    '__cfg_root__',
]


class __PackageInfo(object):

    def __init__(self):
        yaml = YAML(typ='safe')
        yaml.indent(mapping=2, sequence=4, offset=2)
        cwd = Path(__file__).parent / 'conf.yml'
        with open(str(cwd), 'r') as f:
            info = yaml.load(f)
        self._info = info

    def __getattr__(self, item):
        try:
            return self._info[item]
        except KeyError:
            return None


package_info = __PackageInfo()

__pkg_name__ = package_info.name
__version__ = package_info.version
__db_version__ = package_info.db_version
__release__ = package_info.release
__short_description__ = package_info.short_description
__license__ = package_info.license
__author__ = package_info.author
__author_email__ = package_info.author_email
__maintainer__ = package_info.maintainer
__maintainer_email__ = package_info.maintainer_email
__github_username__ = package_info.github_username
__cfg_root__ = str(Path.home() / ('.' + __pkg_name__))



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import re
from collections import defaultdict
from datetime import datetime as dt
from os import remove
from os.path import getmtime
from pathlib import Path
from shutil import rmtree
from warnings import warn

import joblib
import pandas as pd

from ..utils import get_data_loc, absolute_path


class Storage(object):
    """
    Save data in a convenient way:

    .. code:: python

        import numpy as np
        np.random.seed(0)

        # some data
        data1 = np.random.randn(5, 5)
        data2 = np.random.randint(5, 5)

        # init Saver
        saver = LocalStorage('you_dataset_name')

        # save data
        saver(data1, data2)
        # or
        saver.save(data1, data2)

        # retriever data
        date = saver.last()  # last saved
        data = saver[0]  # by index
        for data in saver:  # as iterator
            do_something(data)

        # delete data
        save.rm(0)  # by index
        save.rm()  # delete 'you_dataset_name' dir

    """

    def __init__(self, name=None, *, path=None, mk_dir=True, backend=None):
        """
        Parameters
        ----------
        name: str
            Name of dataset. Usually this is dir name contains data.
        path: str
            Absolute dir path.
        mk_dir: bool
            Ignore ``FileNotFoundError``.
        """
        self._backend = backend or joblib
        self._name = name
        if path is not None:
            self._path = Path(absolute_path(path, mk_dir)) / name
        else:
            self._path = Path(get_data_loc('userdata')) / name
        if not self._path.exists():
            if not mk_dir:
                raise FileNotFoundError()
            self._path.mkdir(parents=True)
        self._files = None
        self._make_file_index()

    @property
    def name(self):
        """
        Dataset name.

        Returns
        -------
        str
        """
        return self._name

    @property
    def path(self):
        """
        Dataset path

        Returns
        -------
        str
        """
        return str(self._path.parent)

    def _make_file_index(self):
        self._files = defaultdict(list)
        files = [f for f in self._path.iterdir() if f.match('*.pkl.*')]

        for f in files:
            # select data
            fn = '.'.join(f.name.split('.')[:-3])

            # for compatibility
            # fixme: will be removed at future
            if fn == 'unnamed':
                warn('file like `unnamed.@x` will be renamed to `@x`.', RuntimeWarning)
                new_name = '.'.join(f.name.split('.')[-3:])
                new_path = f.parent / new_name
                f.rename(new_path)
                f = new_path

            if fn == '':
                fn = 'unnamed'
            self._files[fn].append(f)

        for fs in self._files.values():
            if fs is not None:
                fs.sort(key=lambda f: getmtime(str(f)))

    def _load_data(self, file):
        if file.suffix == '.pd_':
            return pd.read_pickle(str(file))
        else:
            return self._backend.load(str(file))

    def _save_data(self, data, filename):
        self._path.mkdir(parents=True, exist_ok=True)
        if isinstance(data, pd.DataFrame):
            file = self._path / (filename + '.pd.xz')
            pd.to_pickle(data, str(file))
        else:
            file = self._path / (filename + '.pkl.z')
            self._backend.dump(data, str(file))

        return file

    def dump(self, fpath, *, rename=None, with_datetime=True):
        """
        Dump last checked dataset to file.

        Parameters
        ----------
        fpath: str
            Where save to.
        rename: str
            Rename pickle file. Omit to use dataset as name.
        with_datetime: bool
            Suffix file name with dumped time.

        Returns
        -------
        ret: str
            File path.
        """
        ret = {k: self._load_data(v[-1]) for k, v in self._files.items()}
        name = rename if rename else self._name
        if with_datetime:
            datetime = dt.now().strftime('-%Y-%m-%d_%H-%M-%S_%f')
        else:
            datetime = ''
        path_dir = Path(fpath).expanduser()
        if not path_dir.exists():
            path_dir.mkdir(parents=True, exist_ok=True)
        path = path_dir / (name + datetime + '.pkl.z')
        self._backend.dump(ret, str(path))

        return str(path)

    def last(self, name=None):
        """
        Return last saved data.

        Args
        ----
        name: str
            Data's name. Omit for access temp data

        Return
        -------
        ret:any python object
            Data stored in `*.pkl` file.
        """
        try:
            if name is None:
                return self._load_data(self._files['unnamed'][-1])
            if not isinstance(name, str):
                raise TypeError('para `name` must be string but got %s' % type(name))
            return self._load_data(self._files[name][-1])
        except IndexError:
            return None

    def rm(self, index, name=None):
        """
        Delete file(s) with given index.

        Parameters
        ----------
        index: int or slice
            Index of data. Data sorted by datetime.
        name: str
            Data's name. Omit for access unnamed data.
        """

        if not name:
            files = self._files['unnamed'][index]
            if not isinstance(files, list):
                remove(str(files))
            else:
                for f in files:
                    remove(str(f))
            del self._files['unnamed'][index]
            return

        files = self._files[name][index]
        if not isinstance(files, list):
            remove(str(files))
        else:
            for f in files:
                remove(str(f))
        del self._files[name][index]

    def clean(self, name=None):
        """
        Remove all data by name. Omit to remove hole dataset.

        Parameters
        ----------
        name: str
            Data's name.Omit to remove hole dataset.
        """
        if name is None:
            rmtree(str(self._path))
            self._files = list()
            self._files = defaultdict(list)
            return

        for f in self._files[name]:
            remove(str(f))
        del self._files[name]

    def __getattr__(self, name):
        """
        Returns sub-dataset.

        Parameters
        ----------
        name: str
            Dataset name.

        Returns
        -------
        self
        """
        sub_set = Storage(name, path=str(self._path), backend=self._backend)
        setattr(self, name, sub_set)
        return sub_set

    def __repr__(self):
        cont_ls = ['<{}> under `{}` includes:'.format(self._name, self.path)]

        for k, v in self._files.items():
            cont_ls.append('"{}": {}'.format(k, len(v)))

        return '\n'.join(cont_ls)

    def __getitem__(self, item):

        # load file
        def _load_file(files, item_):
            _files = files[item_]
            if not isinstance(_files, list):
                return self._load_data(_files)
            return [self._load_data(f) for f in _files]

        if isinstance(item, tuple):
            try:
                key, index = item
            except ValueError:
                raise ValueError('except 2 parameters. [str, int or slice]')
            if not isinstance(key, str) or \
                    (not isinstance(index, int) and not isinstance(index, slice)):
                raise ValueError('except 2 parameters. [str, int or slice]')
            return _load_file(self._files[key], index)

        if isinstance(item, str):
            return self.__getitem__((item, slice(None, None, None)))

        return _load_file(self._files['unnamed'], item)

    def __call__(self, *unnamed_data, **named_data):
        """
        Same to self.save()
        """
        self.save(*unnamed_data, **named_data)

    def save(self, *unnamed_data, **named_data):
        """
        Save data with or without name.
        Data with same name will not be overwritten.

        Parameters
        ----------
        unnamed_data: any object
            Unnamed data.
        named_data: dict
            Named data as k,v pair.

        """

        def _get_file_index(fn):
            if len(self._files[fn]) != 0:
                return int(re.findall(r'@\d+\.', str(self._files[fn][-1]))[-1][1:-1])
            return 0

        num = 0
        for d in unnamed_data:
            if num == 0:
                num = _get_file_index('unnamed')
            num += 1
            f = self._save_data(d, '@' + str(num))
            self._files['unnamed'].append(f)

        for k, v in named_data.items():
            num = _get_file_index(k) + 1
            f = self._save_data(v, k + '.@' + str(num))
            self._files[k].append(f)



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import warnings

import numpy as np
import torch
from pymatgen.core.structure import Structure

from ..base import BaseGraphFeaturizer
from ...datatools import preset


class CrystalGraphFeaturizer(BaseGraphFeaturizer):

    def __init__(self, *, max_num_nbr=12, radius=8, atom_feature='origin', n_jobs=-1, on_errors='raise',
                 return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``numpy.ndarray`` and ``pandas.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.atom_feature = atom_feature
        self.radius = radius
        self.max_num_nbr = max_num_nbr

    def _atom_feature(self, atom_symbol: str):
        if self.atom_feature == 'origin':
            return preset.atom_init.loc[atom_symbol]
        elif self.atom_feature == 'elements':
            return preset.elements_completed.loc[atom_symbol]
        elif callable(self.atom_feature):
            return self.atom_feature(atom_symbol)
        else:
            raise TypeError('bad `atom feature` parameter')

    def edge_features(self, structure: Structure, **kwargs):
        def expand_distance(distances, dmin=0, step=0.2, var=None):
            """
            Parameters
            ----------
    
            dmin: float
              Minimum interatomic distance
            dmax: float
              Maximum interatomic distance
            step: float
              Step size for the Gaussian filter
            """
            filter_ = np.arange(dmin, self.radius + step, step)
            if var is None:
                var = step

            return np.exp(-(distances[..., np.newaxis] - filter_) ** 2 / var ** 2)

        all_nbrs = structure.get_all_neighbors(self.radius, include_index=True)
        all_nbrs = [sorted(nbrs, key=lambda x: x[1]) for nbrs in all_nbrs]
        nbr_fea_idx, nbr_fea = [], []
        for nbr in all_nbrs:
            if len(nbr) < self.max_num_nbr:
                warnings.warn('can not find enough neighbors to build graph. '
                              'If it happens frequently, consider increase '
                              'radius.')
                nbr_fea_idx.append(list(map(lambda x: x[2], nbr)) +
                                   [0] * (self.max_num_nbr - len(nbr)))
                nbr_fea.append(list(map(lambda x: x[1], nbr)) +
                               [self.radius + 1.] * (self.max_num_nbr -
                                                     len(nbr)))
            else:
                nbr_fea_idx.append(list(map(lambda x: x[2],
                                            nbr[:self.max_num_nbr])))
                nbr_fea.append(list(map(lambda x: x[1],
                                        nbr[:self.max_num_nbr])))
        nbr_fea = np.array(nbr_fea)
        nbr_fea = expand_distance(nbr_fea)
        nbr_fea = torch.Tensor(nbr_fea)

        nbr_fea_idx = torch.LongTensor(nbr_fea_idx)

        return nbr_fea, nbr_fea_idx

    def node_features(self, structure: Structure, **kwargs):
        atom_features = np.vstack([self._atom_feature(s.name) for s in structure.species])
        return torch.Tensor(atom_features)

    @property
    def feature_labels(self):
        return ['atom_feature', 'bond_feature']



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.


__all__ = ['collate_pool', 'CrystalGraphDataset', 'CrystalGraphConvNet', 'ConvLayer']

from .crystal_graph_cnn import collate_pool, CrystalGraphConvNet, ConvLayer, CrystalGraphDataset



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import os
import re
from collections import defaultdict
from pathlib import Path
from warnings import warn

import joblib
import pandas as pd
import requests


class Dataset(object):
    __extension__ = dict(
        pandas=(r'.*\.pd($|\.\w+$)', pd.read_pickle),
        csv=(r'csv', pd.read_csv),
        excel=(r'(xlsx|xls)', pd.read_excel),
        pickle=(r'.*\.pkl($|\.\w+$)', joblib.load))

    __re__ = re.compile(r'[\s\-.]')

    def __init__(self, *paths, backend='pandas', prefix=None):
        self._backend = backend
        self._files = None

        if len(paths) == 0:
            self._paths = ('.',)
        else:
            self._paths = paths

        if not prefix:
            prefix = ()
        self._prefix = prefix

        self._make_index(prefix=prefix)

    def _make_index(self, *, prefix):

        def make(path_):
            def _nest(_f):
                f_ = _f
                return lambda s: s.__extension__[s._backend][1](f_)

            patten = re.compile(self.__extension__[self._backend][0])
            for f in os.listdir(str(path_)):
                if patten.search(f):
                    fn = f.split('.')[0]
                    fn = self.__re__.sub('_', fn)
                    fp = str(path_ / f)

                    # select data
                    parent = re.split(r'[\\/]', str(path_))[-1]
                    # parent = str(f.parent).split('/')[-1]
                    if parent in prefix:
                        fn = '_'.join([parent, fn])

                    if fn in self._files:
                        warn(
                            "file %s with name %s already bind to %s and will be ignored" %
                            (fp, fn, self._files[fn]), RuntimeWarning)
                    else:
                        self._files[fn] = fp
                        setattr(self.__class__, fn, property(_nest(fp)))

        self._files = defaultdict(str)
        for path in self._paths:
            path = Path(path).expanduser().absolute()
            if not path.exists():
                raise RuntimeError('%s not exists' % str(path))
            make(path)

    @classmethod
    def to(cls, obj, path, *, force_pkl=False):
        if isinstance(path, Path):
            path = str(path)
        if not force_pkl and isinstance(obj, (pd.DataFrame, pd.Series)):
            return obj.to_pickle(path)
        return joblib.dump(obj, path)

    @classmethod
    def from_http(cls, url, save_to, *, filename=None, chunk_size=256 * 1024, params=None,
                  **kwargs):
        """
        Get file object via a http request.

        Parameters
        ----------
        url: str
            The resource url.
        save_to: str
            The path of a dir to save the downloaded object into it.
        filename: str, optional
            Specific the file name when saving.
            Set to ``None`` (default) to use a inferred name from http header.
        chunk_size: int, optional
            Chunk size.
        params: any, optional
            Parameters will be passed to ``requests.get`` function.
            See Also: `requests <http://docs.python-requests.org/>`_
        kwargs: dict, optional
            Pass to ``requests.get`` function as the ``kwargs`` parameters.

        Returns
        -------
        str
            File path contains file name.
        """
        r = requests.get(url, params, **kwargs)
        r.raise_for_status()

        if not filename:
            if 'filename' in r.headers:
                filename = r.headers['filename']
            else:
                filename = url.split('/')[-1]

        if isinstance(save_to, str):
            save_to = Path(save_to)
        if not isinstance(save_to, Path) or not save_to.is_dir():
            raise RuntimeError('%s is not a legal path or not point to a dir' % save_to)

        file_ = str(save_to / filename)
        with open(file_, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return file_

    def __repr__(self):
        cont_ls = ['<{}> includes:'.format(self.__class__.__name__)]

        for k, v in self._files.items():
            cont_ls.append('"{}": {}'.format(k, v))

        return '\n'.join(cont_ls)

    @property
    def csv(self):
        return Dataset(*self._paths, backend='csv', prefix=self._prefix)

    @property
    def pandas(self):
        return Dataset(*self._paths, backend='pandas', prefix=self._prefix)

    @property
    def pickle(self):
        return Dataset(*self._paths, backend='pickle', prefix=self._prefix)

    @property
    def excel(self):
        return Dataset(*self._paths, backend='excel', prefix=self._prefix)

    def __call__(self, *args, **kwargs):
        return self.__extension__[self._backend][1](*args, **kwargs)

    def __getattr__(self, name):
        """
        Returns sub-dataset.

        Parameters
        ----------
        name: str
            Dataset name.

        Returns
        -------
        self
        """
        if name in self.__extension__:
            return self.__class__(*self._paths, backend=name, prefix=self._prefix)
        raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from itertools import zip_longest
from pathlib import Path
from warnings import warn

import numpy as np
import pandas as pd
from pymatgen import MPRester
from ruamel.yaml import YAML
from tqdm import tqdm

from .dataset import Dataset
from .._conf import __cfg_root__
from ..utils import config, get_dataset_url, get_sha256, Singleton


class Preset(Dataset, metaclass=Singleton):
    """
    Load data from embed dataset in XenonPy's or user create data saved in ``~/.xenonpy/cached`` dir.
    Also can fetch data by http request.

    This is sample to demonstration how to use is. Also see parameters documents for details.

    ::

        >>> from xenonpy.datatools import preset
        >>> elements = preset.elements
        >>> elements.info()
        <class 'pandas.core.frame.DataFrame'>
        Index: 118 entries, H to Og
        Data columns (total 74 columns):
        atomic_number                    118 non-null int64
        atomic_radius                    88 non-null float64
        atomic_radius_rahm               96 non-null float64
        atomic_volume                    91 non-null float64
        atomic_weight                    118 non-null float64
        boiling_point                    96 non-null float64
        brinell_hardness                 59 non-null float64
        bulk_modulus                     69 non-null float64
        ...
    """

    __dataset__ = ('elements', 'elements_completed', 'atom_init')
    __builder__ = ('mp_samples',)

    # set to check params

    def __init__(self):
        self._dataset = Path(__cfg_root__) / 'dataset'
        self._ext_data = config('ext_data')
        super().__init__(
            str(self._dataset),
            config('userdata'),
            *self._ext_data,
            backend='pandas',
            prefix=('dataset',))

        yaml = YAML(typ='safe')
        yaml.indent(mapping=2, sequence=4, offset=2)

        self._yaml = yaml

    def sync(self, data, to=None):
        """
        load data.

        .. note::
            Try to load data from local at ``~/.xenonpy/dataset``.
            If no data, try to fetch them from remote repository.

        Args
        -----------
        data: str
            name of data.
        to: str
            The version of repository.
            See Also: https://github.com/yoshida-lab/dataset/releases

        Returns
        ------
        ret:DataFrame or Saver or local file path.
        """

        dataset = self._dataset / (data + '.pd.xz')
        sha256_file = self._dataset / 'sha256.yml'

        # check sha256 value
        # make sure sha256_file file exist.
        sha256_file.touch()
        sha256 = self._yaml.load(sha256_file)
        if sha256 is None:
            sha256 = {}

        # fetch data from source if not in local
        if not to:
            url = get_dataset_url(data)
        else:
            url = get_dataset_url(data, to)
        print('fetching dataset `{0}` from {1}.'.format(data, url))
        self.from_http(url, save_to=str(self._dataset))

        sha256_ = get_sha256(str(dataset))
        sha256[data] = sha256_
        self._yaml.dump(sha256, sha256_file)

        self._make_index(prefix=['dataset'])

    def build(self, *keys, save_to=None, **kwargs):

        # build materials project dataset
        def mp_builder(api_key, mp_ids):

            #     print('Will fetch %s inorganic compounds from Materials Project' % len(mp_ids))

            # split requests into fixed number groups
            # eg: grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
            def grouper(iterable, n, fillvalue=None):
                """Collect data into fixed-length chunks or blocks"""
                args = [iter(iterable)] * max(n, 1)
                return zip_longest(fillvalue=fillvalue, *args)

            # the following props will be fetched
            mp_props = [
                'band_gap',
                'density',
                'volume',
                'material_id',
                'pretty_formula',
                'elements',
                'efermi',
                'e_above_hull',
                'formation_energy_per_atom',
                'final_energy_per_atom',
                'unit_cell_formula',
                'structure'
            ]

            entries = []
            mpid_groups = [g for g in grouper(mp_ids, len(mp_ids) // 10)]

            with MPRester(api_key) as mpr:
                for group in tqdm(mpid_groups):
                    mpid_list = [id for id in filter(None, group)]
                    chunk = mpr.query({"material_id": {"$in": mpid_list}}, mp_props)
                    entries.extend(chunk)

            df = pd.DataFrame(entries, index=[e['material_id'] for e in entries])
            df = df.drop('material_id', axis=1)
            df = df.rename(columns={'unit_cell_formula': 'composition'})
            df = df.reindex(columns=sorted(df.columns))

            return df

        for key in keys:
            if key is 'mp_samples':
                if 'api_key' not in kwargs:
                    raise RuntimeError('api key of materials projects database is needed')
                if 'mp_ids' in kwargs:
                    ids = kwargs['mp_ids']
                    if isinstance(ids, (list, tuple)):
                        mp_ids = ids
                    elif isinstance(ids, str):
                        mp_ids = [s.decode('utf-8') for s in np.loadtxt(ids, 'S20')]
                    else:
                        raise ValueError(
                            'parameter `mp_ids` can only be a str to specific the ids file path'
                            'or a list-like object contain the ids')
                else:
                    ids = Path(__file__).absolute().parent / 'mp_ids.txt'
                    mp_ids = [s.decode('utf-8') for s in np.loadtxt(str(ids), 'S20')]
                data = mp_builder(kwargs['api_key'], mp_ids)
                if not save_to:
                    save_to = Path(config('userdata')) / 'mp_samples.pd.xz'
                    save_to = save_to.expanduser().absolute()
                data.to_pickle(save_to)
                self._make_index(prefix=['dataset'])
                return

        raise ValueError('no available key(s) in %s, these can only be %s' % (keys, self.__builder__))

    def _check(self, data):

        dataset = self._dataset / (data + '.pd.xz')
        sha256_file = self._dataset / 'sha256.yml'

        # fetch data from source if not in local
        if not dataset.exists():
            raise RuntimeError(
                "data {0} not exist, please run <Preset.sync('{0}')> to download from the repository".format(data))

        # check sha256 value
        sha256_file.touch()  # make sure sha256_file file exist.
        sha256 = self._yaml.load(sha256_file)
        if sha256 is None:
            sha256 = {}
        if data not in sha256:
            sha256_ = get_sha256(str(dataset))
            sha256[data] = sha256_
            self._yaml.dump(sha256, sha256_file)
        else:
            sha256_ = sha256[data]

        if sha256_ != config(data):
            warn(
                "local version {0} is different from the latest version {1}."
                "you can use <Preset.sync('{0}', to='{1}')> to fix it.".format(data, config('db_version')),
                RuntimeWarning)

    @property
    def elements(self):
        """
        Element properties from embed dataset.
        These properties are summarized from `mendeleev`_, `pymatgen`_, `CRC Handbook`_ and `magpie`_.

        See Also: :doc:`features`

        .. _mendeleev: https://mendeleev.readthedocs.io
        .. _pymatgen: http://pymatgen.org/
        .. _CRC Handbook: http://hbcponline.com/faces/contents/ContentsSearch.xhtml
        .. _magpie: https://bitbucket.org/wolverton/magpie

        Returns
        -------
        DataFrame:
            element properties in pd.DataFrame
        """
        self._check('elements')
        return self.dataset_elements

    @property
    def atom_init(self):
        """
        The initialization vector for each element.

        See Also: https://github.com/txie-93/cgcnn#usage
        """
        self._check('atom_init')
        return self.dataset_atom_init

    @property
    def elements_completed(self):
        """
        Completed element properties. [MICE]_ imputation used

        .. [MICE] `Int J Methods Psychiatr Res. 2011 Mar 1; 20(1): 40–49.`__
                    doi: `10.1002/mpr.329 <10.1002/mpr.329>`_

        .. __: https://www.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&retmode=ref&cmd=prlinks&id=21499542

        See Also: :doc:`features`

        Returns
        -------
            imputed element properties in pd.DataFrame
        """
        self._check('elements_completed')
        return self.dataset_elements_completed


preset = Preset()



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import time
import types
from collections import defaultdict
from contextlib import contextmanager
from datetime import timedelta
from functools import wraps
from os import getenv
from pathlib import Path

from ruamel.yaml import YAML

from ._conf import __cfg_root__, __github_username__, __db_version__


@contextmanager
def set_env(**kwargs):
    """
    Set temp environment variable with ``with`` statement.

    Examples
    --------
    >>> import os
    >>> with set_env(test='test env'):
    >>>    print(os.getenv('test'))
    test env
    >>> print(os.getenv('test'))
    None

    Parameters
    ----------
    kwargs: dict[str]
        Dict with string value.
    """
    import os

    tmp = dict()
    for k, v in kwargs.items():
        tmp[k] = os.getenv(k)
        os.environ[k] = v
    yield
    for k, v in tmp.items():
        if not v:
            del os.environ[k]
        else:
            os.environ[k] = v


def config(key=None, **key_vals):
    """
    Return config value with key or all config.

    Parameters
    ----------
    key: str
        Keys of config item.
    key_vals: dict
        Set item's value by key.
    Returns
    -------
    str
        The value corresponding to the key.
    """
    yaml = YAML(typ='safe')
    yaml.indent(mapping=2, sequence=4, offset=2)
    dir_ = Path(__cfg_root__)
    cfg_file = dir_ / 'conf.yml'

    # from user local
    with open(str(cfg_file), 'r') as f:
        conf = yaml.load(f)

    value = None

    # getter
    if key:
        if key in conf:
            value = conf[key]
        else:
            tmp = Path(__file__).parent / 'conf.yml'
            with open(str(tmp)) as f:
                conf_ = yaml.load(f)

            if key not in conf_:
                raise RuntimeError('No item(s) named %s in configurations' % key)

            value = conf_[key]

    # setter  
    if key_vals:
        for key, v in key_vals.items():
            conf[key] = v
        with open(str(cfg_file), 'w') as f:
            yaml.dump(conf, f)

    return value


def get_dataset_url(name, version=__db_version__):
    """
    Return url with the given file name.

    Args
    ----
    name: str
        binary file name.
    version: str
        The version of repository.
        See Also: https://github.com/yoshida-lab/dataset/releases

    Return
    ------
    str
        binary file url.
    """
    return 'https://github.com/{0}/dataset/releases/download/v{1}/{2}.pd.xz'.format(__github_username__, version,
                                                                                    name)


def get_data_loc(name):
    """Return user data location"""

    scheme = ('userdata', 'usermodel')
    if name not in scheme:
        raise ValueError('{} not in {}'.format(name, scheme))
    if getenv(name):
        return str(Path(getenv(name)).expanduser())
    return str(Path(config(name)).expanduser())


def absolute_path(path, ignore_err=True):
    """
    Resolve path when path include ``~``, ``parent/here``.

    Parameters
    ----------
    path: str
        Path to expand.
    ignore_err: bool
        FileNotFoundError is raised when set to False.
        When True, the path will be created.
    Returns
    -------
    str
        Expanded path.
    """
    from sys import version_info

    if isinstance(path, str):
        path = Path(path)

    if version_info[1] == 5:
        if ignore_err:
            path.expanduser().mkdir(parents=True, exist_ok=True)
        return str(path.expanduser().resolve())
    return str(path.expanduser().resolve(not ignore_err))


def get_sha256(fname):
    """
    Calculate file's sha256 value

    Parameters
    ----------
    fname: str
        File name.

    Returns
    -------
    str
        sha256 value.
    """
    from hashlib import sha256
    hasher = sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(65536), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


class Switch(object):
    def __init__(self, value):
        self.value = value
        self.fall = False

    def __iter__(self):
        """Return the match method once, then stop"""
        yield self.match
        return

    def match(self, *args):
        """Indicate whether or not to enter a case suite"""
        if self.fall or not args:
            return True
        elif self.value in args:  # changed for v1.5, see below
            self.fall = True
            return True
        else:
            return False


class Timer(object):
    class _Timer:
        def __init__(self):
            self.start = None
            self.times = []

        @property
        def elapsed(self):
            all_ = sum(self.times)
            dt = 0.0
            if self.start:
                dt = time.perf_counter() - self.start
            return all_ + dt

        def __repr__(self):
            return 'elapsed: %s' % str(timedelta(seconds=self.elapsed))

    def __init__(self, time_func=time.perf_counter):
        self._func = time_func
        self._timers = defaultdict(self._Timer)

    def __call__(self, *args, **kwargs):
        self.timed(*args, **kwargs)

    def start(self, fn_name='main'):
        if self._timers[fn_name].start is not None:
            raise RuntimeError('Timer <%s> Already started' % fn_name)
        self._timers[fn_name].start = self._func()

    def stop(self, fn_name='main'):
        if self._timers[fn_name].start is None:
            raise RuntimeError('Timer <%s> not started' % fn_name)
        elapsed = self._func() - self._timers[fn_name].start
        self._timers[fn_name].times.append(elapsed)
        self._timers[fn_name].start = None

    @property
    def elapsed(self):
        if 'main' in self._timers:
            return self._timers['main'].elapsed
        return sum([v.elapsed for v in self._timers.values()])

    def __repr__(self):
        return 'Total elapsed: %s\n' % str(timedelta(seconds=self.elapsed)) + \
               '\n'.join(['  |- %s: %s' % (k, v.elapsed) for k, v in self._timers.items()])

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()


class TimedMetaClass(type):
    """
    This metaclass replaces each methods of its classes
    with a new function that is timed
    """

    def _timed(fn):
        if isinstance(fn, (types.FunctionType, types.MethodType)):
            @wraps(fn)
            def fn_(self, *args, **kwargs):
                self._timer.start(fn.__name__)
                try:
                    rt = fn(self, *args, **kwargs)
                finally:
                    self._timer.stop(fn.__name__)
                return rt

            return fn_
        raise TypeError('Need <FunctionType> or <MethodType> but got %s' % type(fn))

    def __new__(mcs, name, bases, attrs):

        if '__init__' in attrs:
            real_init = attrs['__init__']

            # we do a deepcopy in case default is mutable
            # but beware, this might not always work
            @wraps(real_init)
            def injected_init(self, *args, **kwargs):
                setattr(self, '_timer', Timer())
                # call the "real" __init__ that we hid with our injected one
                real_init(self, *args, **kwargs)
        else:
            def injected_init(self):
                setattr(self, '_timer', Timer())
        # inject it
        attrs['__init__'] = injected_init

        for name_, value_ in attrs.items():
            if not name_.startswith("__") and isinstance(value_, (types.FunctionType, types.MethodType)):
                attrs[name_] = TimedMetaClass._timed(value_)

        return super(TimedMetaClass, mcs).__new__(mcs, name, bases, attrs)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.


import numpy as np
from numpy import product


class Product(object):
    def __init__(self, *paras, repeat=1):
        if not isinstance(repeat, int):
            raise ValueError('repeat must be int but got {}'.format(
                type(repeat)))
        lens = [len(p) for p in paras]
        if repeat > 1:
            lens = lens * repeat
        size = product(lens)
        acc_list = [np.floor_divide(size, lens[0])]
        for len_ in lens[1:]:
            acc_list.append(np.floor_divide(acc_list[-1], len_))

        self.paras = paras * repeat if repeat > 1 else paras
        self.lens = lens
        self.size = size
        self.acc_list = acc_list

    def __getitem__(self, index):
        if index > self.size - 1:
            raise IndexError
        ret = [s - 1 for s in self.lens]  # from len to index
        remainder = index + 1
        for i, acc in enumerate(self.acc_list):
            quotient, remainder = np.divmod(remainder, acc)
            if remainder == 0:
                ret[i] = quotient - 1
                break
            ret[i] = quotient

        return tuple(self.paras[i][j] for i, j in enumerate(ret))

    def __len__(self):
        return self.size



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import math
from datetime import datetime, timedelta
from pathlib import Path
from platform import version as sys_ver
from sys import version as py_ver

import numpy as np
import pandas as pd
import torch
import torch.utils.data as Data
from sklearn.base import BaseEstimator

from .checker import Checker
from ..._conf import __version__
from ...utils import TimedMetaClass


class BaseRunner(BaseEstimator, metaclass=TimedMetaClass):

    def __init__(self,
                 epochs=2000,
                 *,
                 cuda: bool or int or str = False,
                 work_dir='.',
                 verbose=True,
                 describe=None):
        self._epochs = epochs
        if isinstance(cuda, bool):
            self._device = torch.device('cuda') if cuda else torch.device('cpu')
        elif isinstance(cuda, int):
            self._device = torch.device('cuda', cuda)
        else:
            self._device = torch.device(cuda)
        if not Path(work_dir).exists():
            Path(work_dir).mkdir()
        self._verbose = verbose
        self._work_dir = work_dir
        self._model = None
        self._model_name = None
        self._checker = None
        self._logs = []
        self._describe = describe or dict(
            python=py_ver,
            system=sys_ver(),
            numpy=np.__version__,
            torch=torch.__version__,
            xenonpy=__version__,
            workspace=work_dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._timer.stop()
        elapsed = str(timedelta(seconds=self.elapsed))
        self.logger('done runner <%s>: %s' % (self.__class__.__name__,
                                              datetime.now().strftime('%Y/%m/%d %H:%M:%S')))
        self.logger('total elapsed time: %s' % elapsed)
        logs = '\n'.join(self._logs)
        now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S_%f')
        with open(str(Path(self._work_dir) / ('log_' + now + '.txt')), 'w') as f:
            f.write(logs)

    def __enter__(self):
        self.logger('start runner <%s> at %s' % (self.__class__.__name__,
                                                 datetime.now().strftime('%Y/%m/%d %H:%M:%S')))
        self._timer.start()
        return self

    def optim(self, iter_):
        """

        Parameters
        ----------
        iter_: generator
            Yields y_train, self._model(x_train), n_ite, n_batch

        Returns
        -------
        any

        """
        raise NotImplementedError

    def post_predict(self, y_true, y_pred):
        """

        Parameters
        ----------
        y_true: torch.Tensor
        y_pred: torch.Tensor

        Returns
        -------
        any
        """
        raise NotImplementedError

    @property
    def epochs(self):
        return self._epochs

    @epochs.setter
    def epochs(self, v):
        self._epochs = v

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, v):
        if not isinstance(v, torch.torch.device):
            raise TypeError('Need torch.device object')
        self._device = v

    @property
    def elapsed(self):
        return self._timer.elapsed

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, m):
        if isinstance(m, torch.nn.Module):
            self._model = m
        else:
            raise TypeError(
                'parameter `m` must be a instance of <torch.nn.modules> but got %s' % type(m))

    def __call__(self, model, name=None, **kwargs):
        """"""
        # model must inherit form nn.Module
        if not isinstance(model, torch.nn.Module):
            raise TypeError('Need <torch.nn.Module> instance but got %s' % type(model))

        if not name:
            from hashlib import md5
            sig = md5(model.__str__().encode()).hexdigest()
            name = sig
        self._checker = Checker(name, self._work_dir)
        self._model = model
        self._model_name = name
        self.describe(model_struct=str(model), model_name=name)
        self._checker.init_model = model
        self._checker.save(
            runner=dict(epochs=self._epochs, verbose=self._verbose, workspace=self._work_dir))

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def describe(self, **info):
        """
        Add some additional description to runner.
        Description will be saved automaticly.

        Parameters
        ----------
        info: dict
            Additional information to describe this model training.
        """
        self._describe = dict(self._describe, **info)

    def persist(self, *args, **kwargs):
        """
        Persist data.
        This is a wrap of ::class:`Dataset`

        Parameters
        ----------
        args
        kwargs
        """
        self._checker.save(*args, **kwargs)

    def checkpoint(self, **describe):
        """
        Take a snapshot for current model status.

        Parameters
        ----------
        describe: dict
            Additional description for checkpoint.

        """
        self._checker(model_state=self._model.state_dict(), **describe)

    @staticmethod
    def tensor(*data_and_type):

        def _tensor(data, torch_type):
            if torch_type is None:
                torch_type = torch.float
            if isinstance(data, pd.DataFrame):
                return torch.from_numpy(data.as_matrix()).to(torch_type)
            elif isinstance(data, np.ndarray):
                return torch.from_numpy(data).to(torch_type)
            else:
                raise TypeError(
                    'Need <numpy.ndarray> or <pandas.DataFrame> but got %s' % type(data))

        return tuple([_tensor(data_, type_) for data_, type_ in data_and_type])

    def to_device(self, *tensor):
        # if use CUDA acc
        if self._device.type != 'cpu':
            return tuple([t.cuda(self._device, True) for t in tensor])
        return tuple([t.cpu() for t in tensor])

    def batch_tensor(self, *data, batch_size=0.2, shuffle=True, num_worker=0, pin_memory=True):
        # batch_size
        if not data:
            return None
        if not all([isinstance(o, (pd.DataFrame, np.ndarray)) for o, _ in data]):
            raise TypeError('Need <numpy.ndarray> or <pandas.DataFrame>')
        if isinstance(batch_size, float):
            batch_size = math.ceil(data[0][0].shape[0] * batch_size)

        return Data.DataLoader(
            dataset=Data.TensorDataset(*self.tensor(*data)),
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_worker,
            pin_memory=pin_memory)

    def logger(self, *info):
        log = '|> ' + '\n|> '.join(info)
        self._logs.append(log)
        if self._verbose:
            print(log)

    def fit(self,
            x_train=None,
            y_train=None,
            *,
            data_loader=None,
            x_dtype=torch.float,
            y_dtype=torch.float):
        """
        Fit Neural Network model

        Parameters
        ----------
        x_train:  ``numpy.ndarray`` or ``pandas.DataFrame``.
            Training data. Will be ignored will``data_loader`` is given.
        y_train:  ``numpy.ndarray`` or ``pandas.DataFrame``.
            Test data. Will be ignored will``data_loader`` is given.
        data_loader: torch.data.DataLoader
            Torch DataLoader. If given, will only use this as training dataset.
        x_dtype: tensor type
            Corresponding dtype in torch tensor. Default is torch.float.
            Detials: https://pytorch.org/docs/stable/tensors.html
        y_dtype: tensor types
            Corresponding dtype in torch tensor. Default is torch.float.

        Returns
        -------
        any
            returns ::meth:`optim` results.
        """

        def _ite():
            self._model.to(self._device)
            if not data_loader:
                x_, y_ = self.to_device(*self.tensor((x_train, x_dtype), (y_train, y_dtype)))
                for t in range(self._epochs):
                    yield y_, self._model(x_), t, 1
                return

            if data_loader:
                for t in range(self._epochs):
                    for i, (x_, y_) in enumerate(data_loader):
                        x_, y_ = self.to_device(x_, y_)
                        yield y_, self._model(x_), t, i
                return

        self.describe(start=datetime.now().strftime('%Y/%m/%d %H:%M:%S'))

        # training
        now = datetime.now().strftime('%Y/%m/%d %H:%M:%S')
        self.logger('training model: <%s>' % self._model_name)
        self.logger('start: %s' % now, '')
        start = self.elapsed

        self._model.train(True)
        ret = self.optim(_ite())  # user implementation
        self._model.train(False)

        now = datetime.now().strftime('%Y/%m/%d %H:%M:%S')
        elapsed = str(timedelta(seconds=self.elapsed - start))
        self.logger('done: %s' % now)
        self.logger('elapsed time: %s\n' % elapsed)

        self.describe(done=now)
        self._checker.save(describe=self._describe)
        self._checker.trained_model = self._model

        return ret

    def predict(self, x_test, y_test=None, *, x_dtype=torch.float, y_dtype=torch.float):
        """

        Parameters
        ----------
        x_test: DataFrame, ndarray
            Input data for test..
        y_test: DataFrame, ndarray, optional
            Target data for test.
        x_dtype: tensor type
            Corresponding dtype in torch tensor. Default is torch.float.
            Detials: https://pytorch.org/docs/stable/tensors.html
        y_dtype: tensor types
            Corresponding dtype in torch tensor. Default is torch.float.

        Returns
        -------
        any
            Return ::meth:`post_predict` results.
        """
        # prepare data
        x_test, = self.to_device(*self.tensor((x_test, x_dtype)))

        # prediction
        self._model.to(self._device)
        y_true, y_pred = y_test, self._model(x_test)
        return self.post_predict(y_true, y_pred)

    @classmethod
    def from_checker(cls, checker, checkpoint=None):
        runner = checker.last('runner')
        ret = cls(
            runner['epochs'],
            work_dir=runner['workspace'],
            verbose=runner['verbose'],
            describe=checker.describe)
        ret._checker = checker
        ret._model_name = checker.model_name
        if not checkpoint:
            ret._model = checker.trained_model if checker.trained_model else checker.init_model
        else:
            model_state, _ = checker[checkpoint]
            ret._model = checker.init_model.load_state_dict(model_state)
        return ret



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

# %%
import json
import os
import tarfile
from pathlib import Path

import numpy as np
import pandas as pd
import requests
from requests import HTTPError
from sklearn.base import BaseEstimator
from tqdm import tqdm

from ..utils import TimedMetaClass, absolute_path


class MDL(BaseEstimator, metaclass=TimedMetaClass):
    def __init__(self, *, save_to='.', api_key=''):
        """
        Access to XenonPy.MDL.

        Parameters
        ----------
        save_to: str
            Path to save models.
            If set to ``None``, only return query result.
        api_key: str
            Not implement yet.
        """
        self.save_to = save_to
        self._url = 'http://xenon.ism.ac.jp/api'
        self.api_key = api_key
        self._headers = {'Accept': 'application/json', "content-type": "application/json"}

    def query(self, query, variables):
        payload = json.dumps({'query': query, 'variables': variables})
        ret = requests.post(url=self._url, headers=self._headers, data=payload)
        if ret.status_code != 200:
            try:
                message = ret.json()
            except json.JSONDecodeError:
                message = "Server did not responce."

            raise HTTPError('status_code: %s, %s' %
                            (ret.status_code, message))
        ret = ret.json()
        if 'errors' in ret:
            raise ValueError(ret['errors'][0]['message'])
        else:
            return ret['data']

    def query_properties(self, name_has):
        query = '''
        query ($name: String!){
          queryProperties(name: $name){
            name
            describe
          }
        }
        '''

        variables = {"name": name_has}
        return self.query(query, variables)['queryProperties']

    def _query_models(self, modelset_has, *,
                      property_has='',
                      descriptor_has='',
                      method_has='',
                      lang_has='',
                      regress_is=True,
                      transferred_is=False,
                      succeed_is=True):
        query = '''
        query ($modelSet: String!
               $property: String
               $descriptor: String
               $method: String
               $lang: String
               $regress: Boolean
               $transferred: Boolean
               $succeed: Boolean)
        {
          queryModels(modelSet: $modelSet
               property: $property
               descriptor: $descriptor
               method: $method
               lang: $lang
               regress: $regress
               transferred: $transferred
               succeed: $succeed)
          {
            mId
            url
            modelSet
            property
            descriptor
            method
            lang
            regress
            transferred
            succeed
            ...on Regression {
              mae
              r
            }
            ...on Classification {
              accuracy
            }
          }
        }
        '''

        variables = dict(
            modelSet=modelset_has,
            property=property_has,
            descriptor=descriptor_has,
            method=method_has,
            lang=lang_has,
            regress=regress_is,
            transferred=transferred_is,
            succeed=succeed_is
        )
        return self.query(query, variables)['queryModels']

    def __call__(self, modelset_has, *,
                 property_has='',
                 descriptor_has='',
                 method_has='',
                 lang_has='',
                 regress_is=True,
                 transferred_is=False,
                 succeed_is=True,
                 **kwargs):
        """
        Query models and download to a specific destination

        Parameters
        ----------
        modelset_has: str
            The part of a model set's name.
            For example, ``modelset_has='test`` will hit ``*test*``
        save_to: str or bool
            Path to save models.
            If ``False``, only return query results.
            This is a temporary change that only have effect in the current fetch.
        property_has: str
            A part of the name of property.
        descriptor_has: str
            A part of the name of descriptor.
        method_has: str
            A part of the name of training method.
        lang_has: str
            A part of the name of programming language.
        regress_is: bool
            If``True``, searching in regression models,
            else, searching in classification models.
            Default is ``True``.
        transferred_is: bool
            If ``True``, searching in transferred models.
            Default is ``False``.
        succeed_is: bool
            If ``True``, searching in succeed models.
            Default is ``True``.

        Returns
        -------
        ret: pd.DataFrame
            A summary of all downloaded models.
        """
        ret = self._query_models(modelset_has,
                                 property_has=property_has,
                                 descriptor_has=descriptor_has,
                                 method_has=method_has,
                                 lang_has=lang_has,
                                 regress_is=regress_is,
                                 transferred_is=transferred_is,
                                 succeed_is=succeed_is)
        if not ret:
            return None
        ret_ = pd.DataFrame(ret)
        if 'save_to' not in kwargs:
            save_to = self.save_to
        else:
            save_to = kwargs['save_to']
        if save_to:
            ret_['save_path'] = self.pull(ret_['url'], save_to=save_to)

        return ret_.set_index('mId', drop=True)

    @classmethod
    def pull(cls, urls, save_to='.'):
        path_list = []
        if isinstance(urls, (np.ndarray, pd.Series)):
            urls = urls.tolist()
        for url in tqdm(urls):
            path = '/'.join(url.split('/')[-5:])
            filename = '{}/{}'.format(absolute_path(save_to), path)
            Path(filename).parent.mkdir(parents=True, exist_ok=True)
            r = requests.get(url=url)
            with open(filename, 'wb') as f:
                f.write(r.content)
            path = filename[:-7]
            tarfile.open(filename).extractall(path=path)
            os.remove(filename)
            path_list.append(path)
        return path_list



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from copy import deepcopy

import numpy as np
from sklearn.base import BaseEstimator

from ..utils import TimedMetaClass


class LogLikelihoodError(Exception):
    """Base exception for LogLikelihood classes"""
    pass


class ResampleError(Exception):
    """Base exception for Resample classes"""
    pass


class ProposalError(Exception):
    """Base exception for Proposal classes"""
    old_smi = None


class SMCError(Exception):
    """Base exception for SMC classes"""
    pass


class BaseLogLikelihood(BaseEstimator, metaclass=TimedMetaClass):

    def fit(self, X, y, **kwargs):
        return self

    def __call__(self, X, **targets):
        return self.log_likelihood(X, **targets)

    def log_likelihood(self, X, **targets):
        """
        Log likelihood

        Parameters
        ----------
        X: list[object]
            Input samples for likelihood calculation.
        targets: tuple[float, float]
            Target area.
            Should be a tuple which have down and up boundary.
            e.g: ``target1=(10, 20)`` equal to ``target1 should in range [10, 20]``.

        Returns
        -------
        log_likelihood: np.ndarray of float
            Estimated likelihood of each samples.
        """
        # raise NotImplementedError('<log_likelihood> have no implementation')
        raise NotImplementedError('<log_likelihood> method must be implemented')


class BaseResample(BaseEstimator, metaclass=TimedMetaClass):

    def fit(self, X, y=None, **kwargs):
        return self

    def __call__(self, X, size, p):
        return self.resample(X, size, p)

    def resample(self, X, size, p):
        """
        Re-sample from given samples.

        Parameters
        ----------
        X: list of object
            Input samples for likelihood calculation.
        size: int
            Resample size.
        p: np.ndarray of float
            The probabilities associated with each entry in X.
            If not given the sample assumes a uniform distribution over all entries.

        Returns
        -------
        new_sample: list of object
            Re-sampling result.
        """
        raise NotImplementedError('<resample> method must be implemented')


class BaseProposal(BaseEstimator, metaclass=TimedMetaClass):
    def fit(self, X, y, **kwargs):
        return self

    def __call__(self, X):
        return self.proposal(X)

    def on_errors(self, error):
        raise error

    def proposal(self, X):
        """
        Proposal new samples based on the input samples.

        Parameters
        ----------
        X: list of object
            Samples for generate next samples

        Returns
        -------
        samples: list of object
            Generated samples from input samples.
        """
        raise NotImplementedError('<proposal> method must be implemented')


class BaseSMC(BaseEstimator, metaclass=TimedMetaClass):
    _log_likelihood = None
    _proposal = None
    _resample = None
    _targets = None

    def log_likelihood(self, X):
        """
        Log likelihood

        Parameters
        ----------
        X: list of object
            Input samples for likelihood calculation.

        Returns
        -------
        log_likelihood: np.ndarray of float
            Estimated likelihood of each samples.
        """
        if self._log_likelihood is None:
            raise NotImplementedError('user need to implement <log_likelihood> method or'
                                      'set <self._log_likelihood> to a instance of <BaseLogLikelihood>')
        return self._log_likelihood(X, **self._targets)

    def resample(self, X, size, p):
        """
        Re-sample from given samples.

        Parameters
        ----------
        X: list[object]
            Input samples for likelihood calculation.
        size: int
            Resample size.
        p: numpy.ndarray[float]
            The probabilities associated with each entry in X.
            If not given the sample assumes a uniform distribution over all entries.

        Returns
        -------
        re-sample: list of object
            Re-sampling result.
        """
        if self._resample is None:
            raise NotImplementedError('user need to implement <resample> method or'
                                      'set <self._resample> to a instance of <BaseResample>')
        return self._resample(X, size, p)

    def proposal(self, X):
        """
        Proposal new samples based on the input samples.

        Parameters
        ----------
        X: list of object
            Samples for generate next samples

        Returns
        -------
        samples: list of object
            Generated samples from input samples.
        """
        if self._proposal is None:
            raise NotImplementedError('user need to implement <proposal> method or'
                                      'set <self._proposal> to a instance of <BaseProposal>')
        return self._proposal(X)

    def on_errors(self, ite, samples, target, error):
        raise error

    def unique(self, X):
        """

        Parameters
        ----------
        X: list of object
            Input samples.

        Returns
        -------
        unique: list of object
            The sorted unique values.
        unique_counts: np.ndarray of int
            The number of times each of the unique values comes up in the original array
        """
        return np.unique(X, return_counts=True)

    @property
    def targets(self):
        return deepcopy(self._targets)

    def update_targets(self, *, reset=False, **targets):
        """
        Update/set the target area.

        Parameters
        ----------
        reset: bool
            If ``true``, reset target area.
        targets: tuple[float, float]
            Target area.
            Should be a tuple which have down and up boundary.
            e.g: ``target1=(10, 20)`` equal to ``target1 should in range [10, 20]``.

        """
        if not self._targets or reset:
            self._targets = {}
        for k, v in targets.items():
            if not isinstance(v, tuple) or len(v) != 2 or v[1] <= v[0]:
                raise ValueError('must be a tuple with (low, up) boundary')
            self._targets[k] = v

    def __setattr__(self, key, value):
        if key is '_log_likelihood' and not isinstance(value, BaseLogLikelihood):
            raise TypeError('<self._log_likelihood> must be a subClass of <BaseLogLikelihood>')
        if key is '_proposal' and not isinstance(value, BaseProposal):
            raise TypeError('<self._proposal> must be a subClass of <BaseProposal>')
        if key is '_resample' and not isinstance(value, BaseResample):
            raise TypeError('<self._resample> must be a subClass of <BaseResample>')
        object.__setattr__(self, key, value)

    def __call__(self, samples, beta, *, size=None, yield_lpf=False, **targets):
        """
        Run SMC

        Parameters
        ----------
        samples: list of object
            Initial samples.
        beta: list of float
            Annealing parameters for each step.
            Should be a 1-D array-like object.
        size: int
            Sample size for each draw.
        yield_lpf : bool
            Yield estimated log likelihood, probability and frequency of each samples. Default is ``False``.
        targets: tuple[float, float]
            Target area.
            Should be a tuple which have down and up boundary.
            e.g: ``target1=(10, 20)`` equal to ``target1 should in range [10, 20]``.

        Yields
        -------
        samples: list of object
            New samples in each SMC iteration.
        llh: np.ndarray float
            Estimated values of log-likelihood of each samples.
            Only yield when ``yield_lpf=Ture``.
        p: np.ndarray of float
            Estimated probabilities of each samples.
            Only yield when ``yield_lpf=Ture``.
        freq: np.ndarray of float
            The number of unique samples in original samples.
            Only yield when ``yield_lpf=Ture``.
        """

        # sample size will be set to the length of init_samples if None
        if size is None:
            size = len(samples)

        if targets:
            self.update_targets(**targets)
        else:
            if not self._targets:
                raise ValueError('must set targets area')

        # translate between input representation and execute environment representation.
        for i, step in enumerate(beta):
            unique, frequency = self.unique(samples)
            try:
                # annealed likelihood in log - adjust with copy counts
                ll = self.log_likelihood(unique)
                w = ll * step + np.log(frequency)
                w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
                p = np.exp(w - w_sum)
                if yield_lpf:
                    yield unique, ll, p, frequency
                else:
                    yield unique

                re_samples = self.resample(unique, size, p)
                samples = self.proposal(re_samples)
            except SMCError as e:
                self.on_errors(i, samples, targets, e)
            except Exception as e:
                raise e

        try:
            unique, frequency = self.unique(samples)
            if yield_lpf:
                ll = self.log_likelihood(unique)
                w = ll + np.log(frequency)
                w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
                p = np.exp(w - w_sum)
                yield unique, ll, p, frequency
            else:
                yield unique
        except SMCError as e:
            self.on_errors(i + 1, samples, targets, e)
        except Exception as e:
            raise e



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import warnings

import numpy as np
import pandas as pd
import torch as tc

from .base import BaseFeaturizer


class FrozenFeaturizer(BaseFeaturizer):
    """
    A Featurizer to extract hidden layers a from NN model.
    """

    def __init__(self, model=None, *, cuda=False, depth=None, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        model: torch.nn.Module
            Source model.
        cuda: bool
            If ``true``, run on GPU.
        depth: int
            The depth will be retrieved from NN model.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """
        super().__init__(n_jobs=0, on_errors=on_errors, return_type=return_type)
        self.depth = depth
        self.model = model
        self.cuda = cuda
        self._ret = []
        self.__authors__ = ['TsumiNa']

    def featurize(self, descriptor, *, depth=None):
        if not isinstance(self.model, tc.nn.Module):
            raise TypeError('<model> must be a instance of <torch.nn.Module>')
        hlayers = []
        if isinstance(descriptor, pd.DataFrame):
            descriptor = descriptor.values
        x_ = tc.from_numpy(descriptor).type(tc.FloatTensor)
        if self.cuda:
            x_.cuda()
            self.model.cuda()
        else:
            x_.cpu()
            self.model.cpu()
        for l in self.model[:-1]:
            x_ = l.layer(x_)
            hlayers.append(x_.data)

        if depth is None:
            depth = self.depth
        if depth is not None:
            ret = hlayers[-depth:]
            if depth > len(ret):
                warnings.warn('<depth> is greater than the max depth of hidden layers')
        else:
            ret = hlayers
            self.depth = len(ret)
        if self.cuda:
            ret = [l.cpu().numpy() for l in ret]
        else:
            ret = [l.numpy() for l in ret]
        self._ret = ret
        return np.concatenate(ret, axis=1)

    @property
    def feature_labels(self):
        depth = len(self._ret)
        if depth == 0:
            raise ValueError('Can not generate labels before transform.')
        return ['L(' + str(i - depth) + ')_' + str(j + 1)
                for i in range(len(self._ret))
                for j in range(self._ret[i].shape[1])]



#  Copyright (c) 2019. stewu5. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import pandas as pd
from mordred import Calculator, descriptors
from rdkit import Chem
from xenonpy.descriptor.base import BaseFeaturizer


class Mordred2DDescriptor(BaseFeaturizer):

    def __init__(self, *, on_errors='raise', return_type='any'):
        # fix n_jobs to be 0 to skip automatic wrapper in XenonPy BaseFeaturizer class
        super().__init__(n_jobs=0, on_errors=on_errors, return_type=return_type)
        self.output = None
        self.__authors__ = ['Stephen Wu', 'TsumiNa']

    def featurize(self, x):
        # check if type(x) = list
        if isinstance(x, pd.Series):
            x = x.tolist()
        if not isinstance(x, list):
            x = [x]
        # check input format, assume SMILES if not RDKit-MOL
        if not isinstance(x[0], Chem.rdchem.Mol):
            x_mol = []
            for z in x:
                x_mol.append(Chem.MolFromSmiles(z))
                if x_mol[-1] is None:
                    raise ValueError('can not convert Mol from SMILES %s' % z)
        else:
            x_mol = x

        calc = Calculator(descriptors, ignore_3D=True)
        self.output = calc.pandas(x_mol)
        return self.output

    @property
    def feature_labels(self):
        return self.output.columns



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

# -*- coding: utf-8 -*-

# uncomment this for compatibility with py27
# from __future__ import print_function

import os
from pathlib import Path

from ruamel.yaml import YAML
from setuptools import setup, find_packages

# YOUR PACKAGE NAME
__package__ = 'xenonpy'


class PackageInfo(object):
    def __init__(self, conf_file):
        yaml = YAML(typ='safe')
        yaml.indent(mapping=2, sequence=4, offset=2)
        with open(conf_file, 'r') as f:
            self._info = yaml.load(f)
        self.name = __package__

    @staticmethod
    def requirements(filename="requirements.txt"):
        """
        Return requirements list from a text file.

        Parameters
        ----------
        filename: str
            Name of requirement file.

        Returns
        -------
            str-list
        """
        try:
            require = list()
            f = open(filename, "rb")
            for line in f.read().decode("utf-8").split("\n"):
                line = line.strip()
                if "#" in line:
                    line = line[:line.find("#")].strip()
                if line:
                    require.append(line)
        except IOError:
            print("'{}' not found!".format(filename))
            require = list()

        return require

    def __getattr__(self, item: str):
        try:
            return self._info[item]
        except KeyError:
            return None


if __name__ == "__main__":
    # --- Automatically generate setup parameters ---
    cwd = Path(__file__).parent / __package__ / 'conf.yml'
    package = PackageInfo(str(cwd))

    # Your package name
    PKG_NAME = package.name

    # Your GitHub user name
    GITHUB_USERNAME = package.github_username

    # Short description will be the description on PyPI
    SHORT_DESCRIPTION = package.short_description  # GitHub Short Description

    # Long description will be the body of content on PyPI page
    LONG_DESCRIPTION = package.long_description

    # Version number, VERY IMPORTANT!
    VERSION = package.version + package.release

    # Author and Maintainer
    AUTHOR = package.author

    # Author email
    AUTHOR_EMAIL = package.author_email

    MAINTAINER = package.maintainer

    MAINTAINER_EMAIL = package.maintainer_email

    PACKAGES, INCLUDE_PACKAGE_DATA, PACKAGE_DATA, PY_MODULES = (
        None,
        None,
        None,
        None,
    )

    # It's a directory style package
    if os.path.exists(__file__[:-8] + PKG_NAME):
        # Include all sub packages in package directory
        PACKAGES = find_packages(
            exclude=["*.tests", "*.tests.*", "tests.*", "tests"])

        # Include everything in package directory
        INCLUDE_PACKAGE_DATA = True
        PACKAGE_DATA = {
            "": ["*.*"],
        }

    # It's a single script style package
    elif os.path.exists(__file__[:-8] + PKG_NAME + ".py"):
        PY_MODULES = [
            PKG_NAME,
        ]

    # Project Url
    GITHUB_URL = "https://github.com/{0}/{1}".format(GITHUB_USERNAME, PKG_NAME)
    # Use todays date as GitHub release tag
    RELEASE_TAG = 'v' + VERSION
    # Source code download url
    DOWNLOAD_URL = "https://github.com/{0}/{1}/archive/{2}.tar.gz".format(
        GITHUB_USERNAME, PKG_NAME, RELEASE_TAG)

    LICENSE = package.license or "'__license__' not found in '%s.__init__.py'!" % PKG_NAME

    PLATFORMS = [
        "Windows",
        "MacOS",
        "Unix",
    ]

    CLASSIFIERS = [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]

    # Read requirements.txt, ignore comments
    INSTALL_REQUIRES = PackageInfo.requirements()
    SETUP_REQUIRES = ['pytest-runner', 'ruamel.yaml']
    TESTS_REQUIRE = ['pytest']
    setup(
        python_requires='~=3.6',
        name=PKG_NAME,
        description=SHORT_DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        packages=PACKAGES,
        include_package_data=INCLUDE_PACKAGE_DATA,
        package_data=PACKAGE_DATA,
        py_modules=PY_MODULES,
        url=GITHUB_URL,
        download_url=DOWNLOAD_URL,
        classifiers=CLASSIFIERS,
        platforms=PLATFORMS,
        license=LICENSE,
        setup_requires=SETUP_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        tests_require=TESTS_REQUIRE)
"""
Appendix
--------
classifiers: https://pypi.python.org/pypi?%3Aaction=list_classifiers
"""



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import numpy as np

from .base import BaseDescriptor, BaseCompositionFeaturizer


class Counting(BaseCompositionFeaturizer):
    def __init__(self, *, one_hot_vec=False, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        one_hot_vec : bool
            Set ``true`` to using one-hot-vector encoding.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.one_hot_vec = one_hot_vec
        self._elems = self._elements.index.tolist()
        self.__authors__ = ['TsumiNa']

    def mix_function(self, elems, nums):
        vec = np.zeros(len(self._elems), dtype=np.int)
        for i, e in enumerate(elems):
            if self.one_hot_vec:
                vec[self._elems.index(e)] = 1
            else:
                vec[self._elems.index(e)] = nums[i]

        return vec

    @property
    def feature_labels(self):
        return self._elems


class WeightedAverage(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = nums / np.sum(nums)
        return w_.dot(elems_)

    @property
    def feature_labels(self):
        return ['ave:' + s for s in self._elements]


class WeightedSum(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = np.array(nums)
        return w_.dot(elems_)

    @property
    def feature_labels(self):
        return ['sum:' + s for s in self._elements]


class GeometricMean(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = np.array(nums).reshape(-1, 1)
        tmp = elems_ ** w_
        return np.power(tmp.prod(axis=0), 1 / sum(w_))

    @property
    def feature_labels(self):
        return ['gmean:' + s for s in self._elements]


class HarmonicMean(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = 1 / self._elements.loc[elems, :].values
        w_ = np.array(nums)
        tmp = w_.dot(elems_)

        return sum(w_) / tmp

    @property
    def feature_labels(self):
        return ['hmean:' + s for s in self._elements]


class WeightedVariance(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, nums):
        elems_ = self._elements.loc[elems, :].values
        w_ = nums / np.sum(nums)
        mean_ = w_.dot(elems_)
        var_ = elems_ - mean_
        return w_.dot(var_ ** 2)

    @property
    def feature_labels(self):
        return ['var:' + s for s in self._elements]


class MaxPooling(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, _):
        elems_ = self._elements.loc[elems, :]
        return elems_.max().values

    @property
    def feature_labels(self):
        return ['max:' + s for s in self._elements]


class MinPooling(BaseCompositionFeaturizer):
    def __init__(self, *, n_jobs=-1, on_errors='raise', return_type='any'):
        """

        Parameters
        ----------
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """

        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)

    def mix_function(self, elems, _):
        elems_ = self._elements.loc[elems, :]
        return elems_.min().values

    @property
    def feature_labels(self):
        return ['min:' + s for s in self._elements]


class Compositions(BaseDescriptor):
    """
    Calculate elemental descriptors from compound's composition.
    """

    def __init__(self, *, n_jobs=-1, featurizers='all', on_errors='nan'):
        """

        Parameters
        ----------
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        n_jobs: int
            The number of jobs to run in parallel for both fit and predict.
            Set -1 to use all cpu cores (default).
            Inputs ``X`` will be split into some blocks then run on each cpu cores.
        featurizers: list[str] or 'all'
            Featurizers that will be used.
            Default is 'all'.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'nan' which will raise up the exception.
        """

        super().__init__(featurizers=featurizers)
        self.n_jobs = n_jobs

        self.composition = Counting(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = WeightedAverage(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = WeightedSum(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = WeightedVariance(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = GeometricMean(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = HarmonicMean(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = MaxPooling(n_jobs=n_jobs, on_errors=on_errors)
        self.composition = MinPooling(n_jobs=n_jobs, on_errors=on_errors)



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.
from functools import partial

import numpy as np
import torch as tc
from scipy.stats import pearsonr
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from torch import nn


class Optim(object):
    @staticmethod
    def sgd(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.SGD`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.SGD
        """
        return partial(tc.optim.SGD, *args, **kwargs)

    @staticmethod
    def ada_delta(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.Adadelta`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.Adadelta
        """
        return partial(tc.optim.Adadelta, *args, **kwargs)

    @staticmethod
    def ada_grad(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.Adagrad`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.Adagrad
        """
        return partial(tc.optim.Adagrad, *args, **kwargs)

    @staticmethod
    def adam(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.Adam`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.Adam
        """
        return partial(tc.optim.Adam, *args, **kwargs)

    @staticmethod
    def sparse_adam(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.SparseAdam`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.SparseAdam
        """
        return partial(tc.optim.SparseAdam, *args, **kwargs)

    @staticmethod
    def ada_max(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.Adamax`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.Adamax
        """
        return partial(tc.optim.Adamax, *args, **kwargs)

    @staticmethod
    def asgd(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.ASGD`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.ASGD
        """
        return partial(tc.optim.ASGD, *args, **kwargs)

    @staticmethod
    def lbfgs(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.LBFGS`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.LBFGS
        """
        return partial(tc.optim.LBFGS, *args, **kwargs)

    @staticmethod
    def rms_prop(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.RMSprop`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.RMSprop
        """
        return partial(tc.optim.RMSprop, *args, **kwargs)

    @staticmethod
    def r_prop(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.Rprop`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.Rprop
        """
        return partial(tc.optim.Rprop, *args, **kwargs)


class LrScheduler(object):
    @staticmethod
    def lambda_lr(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.lr_scheduler.LambdaLR`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.lr_scheduler.LambdaLR
        """
        return partial(tc.optim.lr_scheduler.LambdaLR, *args, **kwargs)

    @staticmethod
    def step_lr(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.lr_scheduler.StepLR`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.lr_scheduler.StepLR
        """
        return partial(tc.optim.lr_scheduler.StepLR, *args, **kwargs)

    @staticmethod
    def multi_step_lr(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.lr_scheduler.MultiStepLR`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.lr_scheduler.MultiStepLR
        """
        return partial(tc.optim.lr_scheduler.MultiStepLR, *args, **kwargs)

    @staticmethod
    def exponential_lr(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.lr_scheduler.ExponentialLR`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.lr_scheduler.ExponentialLR
        """
        return partial(tc.optim.lr_scheduler.ExponentialLR, *args, **kwargs)

    @staticmethod
    def reduce_lr_on_plateau(*args, **kwargs):
        """
        Wrapper class for :class:`torch.optim.lr_scheduler.ReduceLROnPlateau`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau
        """
        return partial(tc.optim.lr_scheduler.ReduceLROnPlateau, *args, **kwargs)


class Init(object):
    @staticmethod
    def uniform(*, scale=0.1):
        b = 1 * scale
        a = -b
        return partial(nn.init.uniform, a=a, b=b)


class L1(object):
    @staticmethod
    def conv(*args, **kwargs):
        """
        Wrapper class for :class:`torch.nn.Conv1d`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.nn.Conv1d
        """
        return partial(nn.Conv1d, *args, **kwargs)

    @staticmethod
    def linear(*args, **kwargs):
        """
        Wrapper class for :class:`torch.nn.Linear`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.nn.Linear
        """
        return partial(nn.Linear, *args, **kwargs)

    @staticmethod
    def batch_norm(*args, **kwargs):
        """
        Wrapper class for :class:`torch.nn.BatchNorm1d`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.nn.BatchNorm1d
        """
        return partial(nn.BatchNorm1d, *args, **kwargs)

    @staticmethod
    def instance_norm(*args, **kwargs):
        """
        Wrapper class for :class:`torch.nn.InstanceNorm1d`.
        http://pytorch.org/docs/0.3.0/optim.html#torch.nn.InstanceNorm1d
        """
        return partial(nn.InstanceNorm1d, *args, **kwargs)


class Metrics(object):
    @staticmethod
    def reg(y_true, y_pred):
        mae = mean_absolute_error(y_true, y_pred)
        rmse = np.sqrt(mean_squared_error(y_true, y_pred))
        r2 = r2_score(y_true, y_pred)
        pr, p_val = pearsonr(y_true, y_pred)
        return dict(
            mae=mae,
            rmse=rmse,
            r2=r2,
            pearsonr=pr,
            p_value=p_val
        )

    # @staticmethod
    # def cls(y_true, y_pred):
    #     mae = mean_absolute_error(y_true, y_pred)
    #     rmse = np.sqrt(mean_squared_error(y_true, y_pred))
    #     r2 = r2_score(y_true, y_pred)
    #     pr, p_val = pearsonr(y_true, y_pred)
    #     return dict(
    #         mae=mae,
    #         rmse=rmse,
    #         r2=r2,
    #         pearsonr=pr,
    #         p_value=p_val
    #     )



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from torch import nn

from .wrap import L1


class Layer1d(nn.Module):
    """
    Base NN layer. This is a wrap around PyTorch.
    See here for details: http://pytorch.org/docs/master/nn.html#
    """

    def __init__(self, n_in, n_out, *,
                 drop_out=0.,
                 layer_func=L1.linear(bias=True),
                 act_func=nn.ReLU(),
                 batch_nor=L1.batch_norm(eps=1e-05, momentum=0.1, affine=True)
                 ):
        """
        Parameters
        ----------
        n_in: int
            Size of each input sample.
        n_out: int
            Size of each output sample
        drop_out: float
            Probability of an element to be zeroed. Default: 0.5
        layer_func: func
            Layers come with PyTorch.
        act_func: func
            Activation function.
        batch_nor: func
            Normalization layers
        """
        super().__init__()
        self.layer = layer_func(n_in, n_out)
        self.batch_nor = None if not batch_nor else batch_nor(n_out)
        self.act_func = None if not act_func else act_func
        self.dropout = None if drop_out == 0. else nn.Dropout(drop_out)

    def forward(self, *x):
        _out = self.layer(*x)
        if self.dropout:
            _out = self.dropout(_out)
        if self.batch_nor:
            _out = self.batch_nor(_out)
        if self.act_func:
            _out = self.act_func(_out)
        return _out



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from functools import partial
from pathlib import Path

import joblib
import torch

from ...datatools.storage import Storage
from ...utils import get_data_loc


class Checker(Storage):
    """
    Check point.
    """

    class __SL(object):
        load = partial(torch.load, map_location=torch.device('cpu'))
        dump = torch.save

    def __init__(self, name, path=None, *, increment=True):
        """
        Parameters
        ----------
        name: str
            Model name.
        path: str
            Save path.
        """
        if not name:
            raise ValueError('need model name.')
        if path is None:
            path = get_data_loc('usermodel')

        if increment:
            i = 1
            while Path(path + '/' + name + '@' + str(i)).exists():
                i += 1
            _fpath = Path(path + '/' + name + '@' + str(i))
        else:
            _fpath = Path(path) / name
        self._name = _fpath.stem
        super().__init__(self._name, path=path)

    @classmethod
    def load(cls, model_path):
        p = Path(model_path)
        return cls(p.stem, str(p.parent), increment=False)

    @property
    def model_name(self):
        return self._name

    @property
    def describe(self):
        """
        Model's description.
        This is a property with getter/setter.
        This action don't overwrite anything but add a new object.

        Returns
        -------
        dict
            Description.
        """
        return self.last('describe')

    @property
    def init_model(self):
        """
        Last appended initial model.
        This is a property with getter/setter.
        This action don't overwrite anything but add a new object.

        Returns
        -------
        model:
            The last appended model.
        """
        self._backend = self.__SL
        try:
            return self.last('init_model')
        except FileNotFoundError:
            return None
        finally:
            self._backend = joblib

    @init_model.setter
    def init_model(self, model):
        """
        Set initial model.
        This action don't overwrite but add a new object.

        Parameters
        ----------
        model: torch.nn.Module
        """
        if isinstance(model, torch.nn.Module):
            self._backend = self.__SL
            super().__call__(init_model=model)
            self._backend = joblib
        else:
            raise TypeError(
                'except `torch.nn.Module` object but got {}'.format(
                    type(model)))

    @property
    def trained_model(self):
        """
        Last appended pre-trained model.
        This is a property with getter/setter.
        This action don't overwrite anything but add a new object.

        Returns
        -------
        model:
            The last appended model.
        """
        self._backend = self.__SL
        try:
            return self.last('trained_model')
        except FileNotFoundError:
            return None
        finally:
            self._backend = joblib

    @trained_model.setter
    def trained_model(self, model):
        """
        Set pre-trained model

        Parameters
        ----------
        model: torch.nn.Module
        """
        if isinstance(model, torch.nn.Module):
            self._backend = self.__SL
            super().__call__(trained_model=model)
            self._backend = joblib
        else:
            raise TypeError(
                'except `torch.nn.Module` object but got {}'.format(
                    type(model)))

    def __getitem__(self, item):
        if isinstance(item, int):
            self._backend = self.__SL
            try:
                cp = self.checkpoints[item]
                model_state = cp['model_state']
                del cp['model_state']
                return model_state, cp
            except IndexError:
                return None
            finally:
                self._backend = joblib
        raise TypeError('except int as checkpoint index but got {}'.format(
            type(item)))

    def __call__(self, **kwargs):
        self._backend = self.__SL
        self.checkpoints(kwargs)
        self._backend = joblib



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

__all__ = ['Compositions', 'Structures', 'Fingerprints']

from .compositions import Compositions, WeightedAverage, WeightedSum, WeightedVariance, \
    MaxPooling, MinPooling, Counting, GeometricMean, HarmonicMean
from .fingerprint import AtomPairFP, TopologicalTorsionFP, RDKitFP, ECFP, FCFP, MACCS, Fingerprints
from .frozen_featurizer import FrozenFeaturizer
from .graph import *
from .structure import Structures, RadialDistributionFunction, ObitalFieldMatrix



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from copy import deepcopy
from types import MethodType
from typing import Union

import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.base import BaseEstimator
from sklearn.linear_model import BayesianRidge

from ..base import BaseLogLikelihood
from ...descriptor.base import BaseDescriptor, BaseFeaturizer


class GaussianLogLikelihood(BaseLogLikelihood):
    def __init__(self, descriptor: Union[BaseFeaturizer, BaseDescriptor], **estimators: BaseEstimator):
        """
        Gaussian loglikelihood.

        Parameters
        ----------
        descriptor: BaseFeaturizer or BaseDescriptor
            Descriptor calculator.
        estimators: BaseEstimator
            Gaussian estimators follow the scikit-learn style.
            These estimators must provide a method named ``predict`` which
            accesses descriptors as input and returns ``(mean, std)`` in order.
            By default, BayesianRidge_ will be used.

            .. _BayesianRidge: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html#sklearn-linear-model-bayesianridge
        """
        if estimators:
            self._mdl = deepcopy(estimators)
        else:
            self._mdl = {}
        if not isinstance(descriptor, (BaseFeaturizer, BaseDescriptor)):
            raise TypeError('<descriptor> must be a subclass of <BaseFeaturizer> or <BaseDescriptor>')
        self._descriptor = descriptor
        self._descriptor.set_params(on_errors='nan')

    def __getitem__(self, item):
        return self._mdl[item]

    def __setitem__(self, key, value):
        if not (hasattr(value, 'predict') and isinstance(value.predict, MethodType)):
            raise TypeError('estimator must be a regressor in scikit-learn style')
        self._mdl[key] = deepcopy(value)

    def remove_estimator(self, *properties: str):
        """
        Remove estimators from estimator set.

        Parameters
        ----------
        properties : str
            The name of properties will be removed from estimator set.
        """
        if not properties:
            self._mdl = {}
        else:
            for p in properties:
                del self._mdl[p]

    def predict(self, smiles, **kwargs):
        fps = self._descriptor.transform(smiles, return_type='df')
        fps_ = fps.dropna()
        tmp = {}
        for k, v in self._mdl.items():
            if isinstance(v, BayesianRidge):
                tmp[k + ': mean'], tmp[k + ': std'] = v.predict(fps_, return_std=True)
            else:
                tmp[k + ': mean'], tmp[k + ': std'] = v.predict(fps_, **kwargs)

        tmp = pd.DataFrame(data=tmp, index=fps_.index)
        return pd.DataFrame(data=tmp, index=fps.index)

    # todo: implement scale function
    def fit(self, smiles, y=None, *, X_scaler=None, y_scaler=None, **kwargs):
        """
        Parameters
        ----------
        smiles: list[str]
            SMILES for training.
        y: pandas.DataFrame
            Target properties for training.
        X_scaler: Scaler (optional, not implement)
            Scaler for transform X.
        y_scaler: Scaler (optional, not implement)
            Scaler for transform y.
        kwargs: dict
            Parameters pass to BayesianRidge initialization.
        """

        if self._mdl:
            raise RuntimeError('estimators have been set.'
                               'If you want to re-train these estimators,'
                               'please use `remove_estimator()` method first.')

        if not isinstance(y, pd.DataFrame):
            raise TypeError('please package all properties into a pd.DataFrame')

        # remove NaN fromm X
        desc = self._descriptor.transform(smiles, return_type='df').reset_index(drop=True)
        y = y.reset_index(drop=True)
        desc.dropna(inplace=True)
        y = y.loc[desc.index]

        for c in y:
            y_ = y[c]  # get target property.
            # remove NaN from y_
            y_.dropna(inplace=True)
            desc_ = desc.loc[y_.index]
            desc_ = desc_.values

            mdl = BayesianRidge(compute_score=True, **kwargs)
            mdl.fit(desc_, y_)
            self._mdl[c] = mdl

    def log_likelihood(self, smis, **targets):
        def _avoid_overflow(ll_):
            # log(exp(log(UP) - log(C)) - exp(log(LOW) - log(C))) + log(C)
            # where C = max(log(UP), max(LOW))
            ll_c = np.max(ll_)
            ll_ = np.log(np.exp(ll_[1] - ll_c) - np.exp(ll_[0] - ll_c)) + ll_c
            return ll_

        ll = np.repeat(-1000.0, len(smis))
        pred = self.predict(smis).reset_index(drop=True)
        tmp = pred.isna().any(axis=1)
        idx = [i for i in range(len(smis)) if ~tmp[i]]

        # calculate likelihood
        ll_mat = []
        for k, (low, up) in targets.items():  # k: target; v: (low, up)

            # predict mean, std for all smiles
            mean, std = pred[k + ': mean'], pred[k + ': std']

            # calculate low likelihood
            low_ll = norm.logcdf(low, loc=np.asarray(mean), scale=np.asarray(std))

            # calculate up likelihood
            up_ll = norm.logcdf(up, loc=np.asarray(mean), scale=np.asarray(std))

            # zip low and up likelihood to a 1-dim array then save it.
            # like: [(tar_low_smi1, tar_up_smi1),  (tar_low_smi2, tar_up_smi2), ..., (tar_low_smiN, tar_up_smiN)]
            lls = zip(low_ll, up_ll)
            ll_mat.append(list(lls))

        # sum all ll along each smiles
        # ll_sum = [[sum_low_smi1, sum_up_smi1], [sum_low_smi2, sum_up_smi2],...,[sum_low_smiN, sum_up_smiN],]
        tmp = [[*map(_avoid_overflow, ll_mat[x])] for x in range(len(ll_mat))]
        ll_sum = np.sum(np.array(tmp), axis=0)

        np.put(ll, idx, ll_sum)
        return ll
